<?php
// $Id: category.inc,v 1.70 2008/07/21 09:30:48 jaza Exp $

/**
 * @file
 * API functions for the category system as provided by the category module.
 * This file is required in order for the category module to work.
 *
 * Other modules should only call functions in this file when interfacing
 * with the category system - functions in category.module itself should not
 * need to be called directly.
 */

/**
 * Return an array of all container objects.
 *
 * @param $type
 *   If set, return only those containers associated with this node type.
 */
function category_get_containers($type = NULL) {
  $type_sql = $type ? " WHERE nt.type = '%s'" : "";
  $result = db_query(db_rewrite_sql("SELECT n.nid, n.title, cn.*, c.weight, c.depth, n.status, nt.type FROM {category_cont} cn INNER JOIN {category} c ON cn.cid = c.cid INNER JOIN {node} n ON cn.cid = n.nid LEFT JOIN {category_cont_node_types} nt ON cn.cid = nt.cid". $type_sql ." ORDER BY c.weight, n.title", 'n', 'nid'), $type);

  $containers = array();
  $node_types = array();
  while ($container = db_fetch_object($result)) {
    // If no node types are associated with a container, the LEFT JOIN will
    // return a NULL value for type.
    if (isset($container->type)) {
      $node_types[$container->cid][$container->type] = $container->type;
      unset($container->type);
      $container->nodes = $node_types[$container->cid];
    }
    elseif (!isset($container->nodes)) {
      $container->nodes = array();
    }

    if (empty($container->admin_title)) {
      $container->admin_title = '';
      $container->has_admin_title = FALSE;
    }
    else {
      $container->has_admin_title = TRUE;
    }
    $containers[$container->cid] = $container;
  }

  return $containers;
}

/**
 * Find all categories associated to the given node, within one container.
 */
function category_node_get_categories_by_container($node, $cnid, $key = 'cid') {
  $result = db_query(db_rewrite_sql('SELECT c.*, n.title FROM {category} c INNER JOIN {category_node} r ON c.cid = r.cid INNER JOIN {node} n ON c.cid = n.nid WHERE c.cnid = %d AND r.vid = %d AND n.status = 1 ORDER BY c.weight'), $cnid, $node->vid);
  $categories = array();
  while ($category = db_fetch_object($result)) {
    $categories[$category->$key] = $category;
  }
  return $categories;
}

/**
 * Find all categories associated to the given node, ordered by container and category weight.
 */
function category_node_get_categories($node, $key = 'cid', $reset = FALSE) {
  static $categories = array();

  if (!isset($categories[$node->vid]) || $reset) {
    $categories[$node->vid] = array();

    $result = db_query(db_rewrite_sql('SELECT n.nid, nr.teaser AS description, r.vid AS node_id, c.*, n.title FROM {category} c INNER JOIN {category_node} r ON c.cid = r.cid INNER JOIN {category} cn ON c.cnid = cn.cid INNER JOIN {node} n ON c.cid = n.nid INNER JOIN {node_revisions} nr ON n.vid = nr.vid INNER JOIN {node} cnn ON cn.cid = cnn.nid WHERE n.status = 1 AND r.vid = %d ORDER BY cn.weight, cnn.title, c.weight, n.title', 'n', 'nid'), $node->vid);

    while ($category = db_fetch_object($result)) {
      $categories[$node->vid][$category->$key] = $category;
    }
  }

  return $categories[$node->vid];
}

/**
 * Find all category objects related to a given category ID.
 */
function category_get_related($cid, $key = 'cid') {
  if ($cid) {
    $result = db_query(db_rewrite_sql('SELECT c.*, n.title, cid1, cid2 FROM {category_relation}, {category} c, {node} n WHERE (c.cid = cid1 OR c.cid = cid2) AND (cid1 = %d OR cid2 = %d) AND c.cid = n.nid AND c.cid != %d AND n.status = 1 ORDER BY c.weight, n.title'), $cid, $cid, $cid);
    $related = array();
    while ($category = db_fetch_object($result)) {
      $related[$category->$key] = $category;
    }
    return $related;
  }
  else {
    return array();
  }
}

/**
 * Find all parents of a given category ID.
 * This function will return the parents in an indexed array sorted by the parents' weight and title.
 */
function category_get_parents($cid, $key = 'cid', $distant = TRUE, $reset = FALSE) {
  static $parents, $distant_parents;

  if (!$cid) {
    return array();
  }

  if (($distant && !isset($distant_parents)) || (!$distant && !isset($parents)) || $reset) {
    $distant_sql = $distant ? '' : ' AND (c.cnid = cc.cnid OR c.cnid = 0)';
    $result = db_query(db_rewrite_sql('SELECT c.*, h.cid AS child, n.title, cn.admin_title FROM {category} c INNER JOIN {category_hierarchy} h ON c.cid = h.parent INNER JOIN {node} n ON c.cid = n.nid LEFT JOIN {category_cont} cn ON c.cid = cn.cid INNER JOIN {category} cc ON h.cid = cc.cid WHERE n.status = 1'. $distant_sql .' ORDER BY c.weight, n.title'), $cid);
    if ($distant) {
      $distant_parents = array();
    }
    else {
      $parents = array();
    }
    while ($parent = db_fetch_object($result)) {
      if (empty($parent->admin_title)) {
        $parent->admin_title = $parent->title;
      }
      if ($distant) {
        $distant_parents[$parent->child][$parent->$key] = $parent;
      }
      else {
        $parents[$parent->child][$parent->$key] = $parent;
      }
    }
  }

  $no_parent = new StdClass();
  $no_parent->cid = 0;
  $no_parent->cnid = 0;
  $no_parent->weight = 0;
  $no_parent->depth = 0;
  $no_parent->title = '';
  $no_parent->admin_title = '';
  $no_parent->child = $cid;
  return $distant ? (isset($distant_parents[$cid]) ? $distant_parents[$cid] : array(0 => $no_parent)) : (isset($parents[$cid]) ? $parents[$cid] : array(0 => $no_parent));
}

/**
 * Find all ancestors of a given category ID.
 */
function category_get_parents_all($cid, $distant = FALSE) {
  $parents = array();
  if ($cid) {
    $parents[] = category_get_category($cid);
    $n = 0;
    while ($parent = category_get_parents($parents[$n]->cid, 'cid', $distant)) {
      $parents = array_merge($parents, $parent);
      $n++;
    }
  }
  return $parents;
}

/**
 * Find all children of a category ID.
 */
function category_get_children($cid, $cnid = 0, $key = 'cid', $reset = FALSE) {
  static $children;

  if ($cnid) {
    $result = db_query(db_rewrite_sql('SELECT c.*, n.title FROM {category} c INNER JOIN {category_hierarchy} h ON c.cid = h.cid INNER JOIN {node} n ON c.cid = n.nid WHERE h.parent = %d AND n.status = 1 AND c.cnid = %d ORDER BY c.weight, n.title'), $cid, $cnid);

    $results = array();
    while ($category = db_fetch_object($result)) {
      $results[$category->$key] = $category;
    }

    return $results;
  }
  elseif (!isset($children) || $reset) {
    $result = db_query(db_rewrite_sql('SELECT c.*, h.parent, n.title FROM {category} c INNER JOIN {category_hierarchy} h ON c.cid = h.cid INNER JOIN {node} n ON c.cid = n.nid WHERE n.status = 1 ORDER BY c.weight, n.title'), $cid);

    $children = array();
    while ($category = db_fetch_object($result)) {
      $children[$category->parent][$category->$key] = $category;
    }
  }

  return isset($children[$cid]) ? $children[$cid] : array();
}

/**
 * Create a hierarchical representation of a set of categories.
 *
 * @param $cnid
 *   Which container to generate the tree for.
 *
 * @param $parent
 *   The node ID under which to generate the tree. If 0, generate the tree
 *   for everything under the container.
 *
 * @param $depth
 *   Internal use only.
 *
 * @param $max_depth
 *   The number of levels of the tree to return. Leave NULL to return all levels.
 *
 * @param $distant
 *   Whether other containers or categories under them should be returned as
 *   part of the tree. Default is false.
 *
 * @return
 *   An array of all category objects in the tree. Each category object is extended
 *   to have "depth" and "parents" attributes in addition to its normal ones.
 */
function category_get_tree($cnid, $parent = NULL, $depth = -1, $max_depth = NULL, $distant = FALSE) {
  static $children, $parents, $categories;

  if (empty($cnid) && $depth == -1) {
    $distant = TRUE;
  }
  if (!isset($parent)) {
    $parent = $cnid;
  }

  $depth++;

  // We cache trees, so it's not CPU-intensive to call get_tree() on a category
  // and its children, too.
  if (!isset($children[$cnid])) {
    $distant_sql = ($distant) ? '' : 'AND c.cnid = %d ';

    $children[$cnid] = array();
    $result = db_query(db_rewrite_sql('SELECT c.cid, c.cnid, c.weight, c.depth, h.parent, n.title, cn.admin_title FROM {category} c INNER JOIN {category_hierarchy} h ON c.cid = h.cid INNER JOIN {node} n ON c.cid = n.nid LEFT JOIN {category_cont} cn ON c.cid = cn.cid LEFT JOIN {category} c2 ON h.parent = c2.cid WHERE (c2.cid IS NOT NULL OR h.parent = 0) AND n.status = 1 '. $distant_sql .'ORDER BY c.weight, n.title', 'n', 'nid'), $cnid);

    while ($category = db_fetch_object($result)) {
      if (!$distant && $depth == 0) {
        $parent_cat = category_get_category($category->parent);
        if ($parent_cat->cid != $cnid && $parent_cat->cnid != $cnid) {
          $category->parent = $cnid;
        }
      }

      $children[$cnid][$category->parent][] = $category->cid;
      $categories[$cnid][$category->cid] = $category;
      $parents[$cnid][$category->cid][] = $category->parent;
    }
  }

  $max_depth = (is_null($max_depth) && isset($children[$cnid])) ? count($children[$cnid]) : $max_depth;
  $tree = array();
  if (!empty($children[$cnid]) && !empty($children[$cnid][$parent])) {
    foreach ($children[$cnid][$parent] as $child) {
      if ($max_depth > $depth) {
        $cat = drupal_clone($categories[$cnid][$child]);
        $cat->depth = $depth;
        // The "parent" attribute is not useful, as it would show one parent only.
        unset($cat->parent);
        $cat->parents = $parents[$cnid][$child];

        if (!$cat->cnid && empty($cat->admin_title)) {
          $cat->admin_title = $cat->title;
        }
        $tree[] = $cat;

        if (!empty($children[$cnid][$child])) {
          $tree = array_merge($tree, category_get_tree($cnid, $child, $depth, $max_depth, $distant));
        }
      }
    }
  }

  return $tree;
}

/**
 * Return an array of synonyms of the given category ID.
 */
function category_get_synonyms($cid) {
  if ($cid) {
    $synonyms = array();
    $result = db_query('SELECT name FROM {category_synonym} WHERE cid = %d', $cid);
    while ($synonym = db_fetch_array($result)) {
      $synonyms[] = $synonym['name'];
    }
    return $synonyms;
  }
  else {
    return array();
  }
}

/**
 * Return the category object that has the same given string as a synonym.
 */
function category_get_synonym_root($synonym) {
  return db_fetch_object(db_query("SELECT c.*, n.title FROM {category} c INNER JOIN {category_synonym} s ON c.cid = s.cid AND s.name = '%s' INNER JOIN {node} n ON c.cid = n.nid", $synonym));
}

/**
 * Given a category id, count the number of published nodes in it.
 */
function category_category_count_nodes($cid, $type = 0) {
  static $count;

  if (!isset($count[$type])) {
    // $type == 0 always evaluates true if $type is a string
    if (is_numeric($type)) {
      $result = db_query(db_rewrite_sql('SELECT c.cid, COUNT(n.nid) AS cnt FROM {category_node} c INNER JOIN {node} n ON c.nid = n.nid WHERE n.status = 1 GROUP BY c.cid'));
    }
    else {
      $result = db_query(db_rewrite_sql("SELECT c.cid, COUNT(n.nid) AS cnt FROM {category_node} c INNER JOIN {node} n ON c.nid = n.nid WHERE n.status = 1 AND n.type = '%s' GROUP BY c.cid"), $type);
    }
    while ($category = db_fetch_object($result)) {
      $count[$type][$category->cid] = $category->cnt;
    }
  }
  $children_count = 0;
  foreach (_category_category_children($cid) as $c) {
    $children_count += category_category_count_nodes($c, $type);
  }
  return $children_count + (isset($count[$type][$cid]) ? $count[$type][$cid] : 0);
}

/**
 * Helper for category_category_count_nodes().
 */
function _category_category_children($cid) {
  static $children;

  if (!isset($children)) {
    $result = db_query('SELECT cid, parent FROM {category_hierarchy}');
    while ($category = db_fetch_object($result)) {
      $children[$category->parent][] = $category->cid;
    }
  }
  return isset($children[$cid]) ? $children[$cid] : array();
}

/**
 * Try to map a string to an existing category, as for glossary use.
 *
 * Provides a case-insensitive and trimmed mapping, to maximize the
 * likelihood of a successful match.
 *
 * @param title
 *   Title of the category to search for.
 *
 * @return
 *   An array of matching category objects.
 */
function category_get_category_by_name($title) {
  $db_result = db_query(db_rewrite_sql("SELECT n.nid, n.title, c.* FROM {category} c INNER JOIN {node} n ON c.cid = n.nid WHERE LOWER('%s') LIKE LOWER(n.title)", 'n', 'nid'), trim($title));
  $result = array();
  while ($category = db_fetch_object($db_result)) {
    $result[] = $category;
  }

  return $result;
}

/**
 * Return the container object matching a container ID.
 *
 * @param $cnid
 *   The container's ID
 *
 * @return
 *   The container object with all of its metadata, if exists, NULL otherwise.
 *   Results are statically cached.
 */
function category_get_container($cnid) {
  static $containers = array();

  if (!isset($containers[$cnid])) {
    // Initialize so if this container does not exist, we have
    // that cached, and we will not try to load this later.
    $containers[$cnid] = FALSE;
    // Try to load the data and fill up the object.
    $result = db_query(db_rewrite_sql('SELECT n.nid, cn.*, n.title, c.weight, c.depth, ct.type, c.weight AS cont_weight, c.depth AS cont_depth FROM {category} c INNER JOIN {category_cont} cn ON c.cid = cn.cid INNER JOIN {node} n ON c.cid = n.nid LEFT JOIN {category_cont_node_types} ct ON c.cid = ct.cid WHERE c.cid = %d ORDER BY c.weight, n.title', 'n', 'nid'), $cnid);
    $node_types = array();

    while ($cont = db_fetch_object($result)) {
      // Find all allowed node types for this container.
      if (!empty($cont->type)) {
        $node_types[$cont->type] = $cont->type;
      }
      unset($cont->type);
      $cont->nodes = $node_types;

      // If no admin title is specified, inherit the container title.
      if (empty($cont->admin_title)) {
        $cont->admin_title = '';
        $cont->has_admin_title = FALSE;
      }
      else {
        $cont->has_admin_title = TRUE;
      }

      $containers[$cnid] = $cont;
    }
  }

  // Return NULL if this vocabulary does not exist.
  return (isset($containers[$cnid]) ? $containers[$cnid] : NULL);
}

/**
 * Retrieve the category object matching a category ID.
 *
 * @param $cid
 *   The node/category ID of the category to be retrieved.
 * @param $reset
 *   Whether or not to reset the cache of categories. Default is FALSE.
 *
 * @return
 *   Populated category object if the category is found, NULL otherwise.
 */
function category_get_category($cid, $reset = FALSE) {
  static $categories = array();

  if (!isset($categories[$cid])) {
    $categories[$cid] = db_fetch_object(db_query('SELECT n.nid, n.title, c.* FROM {category} c INNER JOIN {node} n ON c.cid = n.nid WHERE cid = %d', $cid));
  }

  return (isset($categories[$cid]) ? $categories[$cid] : NULL);
}

/**
 * Finds all nodes that match selected category conditions.
 *
 * @param $cids
 *   An array of category IDs to match.
 * @param $operator
 *   How to interpret multiple IDs in the array. Can be "or" or "and".
 * @param $depth
 *   How many levels deep to traverse the category tree. Can be a nonnegative
 *   integer or "all".
 * @param $pager
 *   Whether the nodes are to be used with a pager (the case on most Drupal
 *   pages) or not (in an XML feed, for example).
 * @param $distant
 *   Whether other containers or categories under them should be returned as
 *   part of the tree. Default is false.
 * @param $order
 *   The order clause for the query that retrieve the nodes.
 * @return
 *   A resource identifier pointing to the query results.
 */
function category_select_nodes($cids = array(), $operator = 'or', $depth = 0, $pager = TRUE, $distant = FALSE, $order = 'n.sticky DESC, n.created DESC') {
  if (count($cids) > 0) {
    // For each category ID, generate an array of descendant category IDs to the right depth.
    $descendant_cids = array();
    if ($depth === 'all' || $depth < 0) {
      $depth = NULL;
    }
    foreach ($cids as $index => $cid) {
      $category = category_get_category($cid);
      $cnid = $category->cnid ? $category->cnid : $cid;
      $tree = category_get_tree($cnid, $cid, -1, $depth, $distant);
      $descendant_cids[] = array_merge(array($cid), array_map('_category_get_cid_from_category', $tree));
    }

    $args = array();
    if ($operator == 'or') {
      $args = call_user_func_array('array_merge', $descendant_cids);
      $placeholders = db_placeholders($args, 'int');
      $sql = 'SELECT DISTINCT(n.nid), n.sticky, n.title, n.created FROM {node} n INNER JOIN {category_node} cn ON n.nid = cn.nid WHERE cn.cid IN ('. $placeholders .') AND n.status = 1 ORDER BY '. $order;
      $sql_count = 'SELECT COUNT(DISTINCT(n.nid)) FROM {node} n INNER JOIN {category_node} cn ON n.nid = cn.nid WHERE cn.cid IN ('. $placeholders .') AND n.status = 1';
    }
    else {
      $joins = '';
      $wheres = '';
      foreach ($descendant_cids as $index => $cids) {
        $joins .= ' INNER JOIN {category_node} cn'. $index .' ON n.nid = cn'. $index .'.nid';
        $wheres .= ' AND cn'. $index .'.cid IN ('. db_placeholders($cids, 'int') .')';
        $args = array_merge($args, $cids);
      }
      $sql = 'SELECT DISTINCT(n.nid), n.sticky, n.title, n.created FROM {node} n '. $joins .' WHERE n.status = 1 '. $wheres .' ORDER BY '. $order;
      $sql_count = 'SELECT COUNT(n.nid) FROM {node} n '. $joins .' WHERE n.status = 1 ' . $wheres;
    }
    $sql = db_rewrite_sql($sql);
    $sql_count = db_rewrite_sql($sql_count);
    if ($pager) {
      $result = pager_query($sql, variable_get('default_nodes_main', 10), 0, $sql_count, $args);
    }
    else {
      $result = db_query_range($sql, $args, 0, variable_get('feed_default_items', 10));
    }
  }

  return $result;
}

/**
 * Accepts the result of a pager_query() call, such as that performed by
 * category_select_nodes(), and formats each node along with a pager.
*/
function category_render_nodes($result, $category_display) {
  $output = '';
  $has_rows = FALSE;
  while ($node = db_fetch_object($result)) {
    $output .= node_view(node_load($node->nid), 1);
    $has_rows = TRUE;
  }
  if ($has_rows) {
    $output .= theme('pager', NULL, variable_get('default_nodes_main', 10), 0);
  }
  else {
    if (!module_exists('category_display') || !empty($category_display->emptymsg)) {
      $output .= '<p>'. t('There are currently no posts in this category.') .'</p>';
    }
  }

  return $output;
}

/**
 * Parses a comma or plus separated string of category IDs.
 *
 * @param $str_cids
 *   A string of category IDs, separated by plus or comma.
 *   comma (,) means AND
 *   plus (+) means OR
 *
 * @return an associative array with an operator key (either 'and'
 *   or 'or') and a cid key containing an array of the category IDs.
 */
function category_categories_parse_string($str_cids) {
  $categories = array('operator' => '', 'cids' => array());
  if (preg_match('/^([0-9]+[+ ])+[0-9]+$/', $str_cids)) {
    $categories['operator'] = 'or';
    // The '+' character in a query string may be parsed as ' '.
    $categories['cids'] = preg_split('/[+ ]/', $str_cids);
  }
  else if (preg_match('/^([0-9]+,)*[0-9]+$/', $str_cids)) {
    $categories['operator'] = 'and';
    $categories['cids'] = explode(',', $str_cids);
  }
  return $categories;
}
